@@ -13,4 +13,4 @@ if __name__ == "__main__":
 
     # Preprocessing
     preprocessor = DataPreprocessor(Settings().data_config_filename, args.rules)
-    preprocessor.preprocess(nproc=args.nproc, batch_size=args.batchsize)
+    preprocessor.preprocess(nproc=args.nproc)
