@@ -7,6 +7,7 @@ Module retrieving text templates
 # Standard packages
 from __future__ import annotations
 import json
+from pathlib import Path
 from typing import Any, Callable, Hashable, Mapping, Optional, Sequence, Tuple, Union
 from configparser import ConfigParser
 from copy import deepcopy
@@ -14,14 +15,16 @@ from copy import deepcopy
 # Third parties packages
 import pandas as pd
 import numpy as np
+from dtaidistance import dtw
 from mfire.settings import get_logger
+from mfire.utils import FormatDict
 
 # Logging
 LOGGER = get_logger(name="text.template.mod", bind="text.template")
 
 
 def read_file(
-    filename: str,
+    filename: Path,
     col: Optional[str] = "template",
     weights: Optional[np.ndarray] = None,
     **kwargs,
@@ -30,7 +33,7 @@ def read_file(
     TemplateRetriever object.
 
     Args:
-        filename (str): Name of the file containing the templates.
+        filename (Path): Name of the file containing the templates.
         col (Optional[str]): templates column name in a csv file.
         weights (Optional[np.ndarray]): weights if centroids.
         **kwargs (Optional[dict]) : keyword arguments of the file reader used :
@@ -41,11 +44,12 @@ def read_file(
     Returns:
         TemplateRetriever: TemplateRetriever object.
     """
-    if filename.endswith(".json"):
+    filename = Path(filename)
+    if filename.suffix == ".json":
         return JsonTemplateRetriever.read_file(filename, **kwargs)
-    if filename.endswith(".ini"):
+    if filename.suffix == ".ini":
         return IniTemplateRetriever.read_file(filename, **kwargs)
-    if filename.endswith(".csv"):
+    if filename.suffix == ".csv":
         if weights is not None:
             return CentroidTemplateRetriever.read_file(
                 filename, col=col, weights=weights, **kwargs
@@ -63,37 +67,26 @@ def read_file(
     return TemplateRetriever.read_file(filename, **kwargs)
 
 
-class FormatDict(dict):
-    """FormatDict: dictionnary extension for handling string formatting
-    with missing keys. For instance :
-    >>> dico = {"key1" : "value1"}
-    >>> "la valeur 1 = {key1}, la valeur 2 = {key2}".format_map(dico)
-    ...
-    KeyError: "value2" missing
-    >>> dico = FormatDict(dico)
-    >>> dico
-    {"key1" : "value1"}
-    >>> "la valeur 1 = {key1}, la valeur 2 = {key2}".format_map(dico)
-    "la valeur 1 = value1, la valeur 2 = {key2}"
-
-    Inheritance:
-        dict
-    """
-
-    def __missing__(self, key):
-        return "{" + key + "}"
-
-
 class Template(str):
     """Template: class extending str type for formating using a custom FormatDict
     (in order to avoid issues when keys are missing).
     >>> s = Template("Ma {b} connait des blagues de {a}, et {c} ?")
     >>> s.format(a="toto", b="tata")
     "Ma tata connait des blagues de toto, et {c} ?"
+    >>> s = Template("Bonjour M. {a_prenom} {a_nom}.")
+    >>> s.format(a={"nom": "dupont", "prenom": "toto"}, b="tata")
+    "Bonjour M. toto dupont."
     """
 
     def format(self, **kwargs) -> Template:
-        return self.format_map(FormatDict(kwargs))
+        try:
+            flat_kwargs = pd.json_normalize(kwargs, sep="_").to_dict(orient="records")[
+                0
+            ]
+        except IndexError:
+            flat_kwargs = kwargs
+
+        return self.format_map(FormatDict(flat_kwargs))
 
 
 class TemplateRetriever:
@@ -117,17 +110,17 @@ class TemplateRetriever:
         return f"{self.__class__}\n{self.table}"
 
     @classmethod
-    def read_file(cls, filename: str, **kwargs) -> TemplateRetriever:
+    def read_file(cls, filename: Path, **kwargs) -> TemplateRetriever:
         """read_file : class method for instantiating an TemplateRetriever
         out of a file.
 
         Args:
-            filename (str): File's name
+            filename (Path): File's name
 
         Returns:
             TemplateRetriever: New cls object
         """
-        table = {"filename": filename}
+        table = {"filename": Path(filename)}
         table.update(kwargs)
         return cls(table)
 
@@ -174,17 +167,17 @@ class TemplateRetriever:
 class JsonTemplateRetriever(TemplateRetriever):
     """JsonTemplateRetriever : TemplateRetriever from json files.
 
-        Currently used by:
-            - mfire.text.comment.comment_builder
+    Currently used by:
+        - mfire.text.comment.comment_builder
     """
 
     @classmethod
-    def read_file(cls, filename: str, **kwargs) -> JsonTemplateRetriever:
+    def read_file(cls, filename: Path, **kwargs) -> JsonTemplateRetriever:
         """read_file : class method for instantiating an JsonTemplateRetriever
         out of a file.
 
         Args:
-            filename (str): File's name
+            filename (Path): File's name
             **kwargs (Optional[dict]) : keyword arguments of the file reader used
                 (json.load)
 
@@ -196,7 +189,7 @@ class JsonTemplateRetriever(TemplateRetriever):
                 return cls(json.load(json_file, **kwargs))
         except Exception:
             LOGGER.error(f"Failed to read json template file {filename}", exc_info=True)
-        return cls({"filename": filename})
+        return cls({"filename": Path(filename)})
 
     def get(
         self,
@@ -253,16 +246,16 @@ class JsonTemplateRetriever(TemplateRetriever):
 class IniTemplateRetriever(TemplateRetriever):
     """JsonTemplateRetriever : TemplateRetriever from ini files.
 
-        Currently unused.
+    Currently unused.
     """
 
     @classmethod
-    def read_file(cls, filename: str, **kwargs) -> IniTemplateRetriever:
+    def read_file(cls, filename: Path, **kwargs) -> IniTemplateRetriever:
         """read_file : class method for instantiating an IniTemplateRetriever
         out of a file.
 
         Args:
-            filename (str): File's name
+            filename (Path): File's name
             **kwargs (Optional[dict]) : keyword arguments of the file reader used
                 (ConfigParser().read())
 
@@ -279,10 +272,12 @@ class IniTemplateRetriever(TemplateRetriever):
                 filename=filename,
                 exc_info=True,
             )
-            return cls({"filename": filename})
+            return cls({"filename": Path(filename)})
 
     def get(
-        self, key: Union[str, Tuple[str, str]], default: Optional[str] = None,
+        self,
+        key: Union[str, Tuple[str, str]],
+        default: Optional[str] = None,
     ) -> str:
         """get: method used to retrieve a specific template identified by a key.
 
@@ -316,7 +311,7 @@ class IniTemplateRetriever(TemplateRetriever):
 class CsvTemplateRetriever(TemplateRetriever):
     """CsvTemplateRetriever : TemplateRetriever for csv files.
 
-        Currently unused.
+    Currently unused.
     """
 
     def __init__(
@@ -345,12 +340,12 @@ class CsvTemplateRetriever(TemplateRetriever):
         self.table = self.table[self.col]
 
     @classmethod
-    def read_file(cls, filename: str, **kwargs) -> CsvTemplateRetriever:
+    def read_file(cls, filename: Path, **kwargs) -> CsvTemplateRetriever:
         """read_file : class method for instantiating an CsvTemplateRetriever
         out of a file.
 
         Args:
-            filename (str): File's name
+            filename (Path): File's name
             **kwargs (Optional[dict]) : keyword arguments of the file reader used
                 (pandas.read_csv)
 
@@ -366,9 +361,13 @@ class CsvTemplateRetriever(TemplateRetriever):
                 filename=filename,
                 exc_info=True,
             )
-            return cls({"filename": filename})
+            return cls({"filename": Path(filename)})
 
-    def get(self, key: Union[Hashable, Tuple], default: Optional[str] = None,) -> str:
+    def get(
+        self,
+        key: Union[Hashable, Tuple],
+        default: Optional[str] = None,
+    ) -> str:
         """get: method used to retrieve a specific template identified by a key.
 
         Args:
@@ -383,7 +382,10 @@ class CsvTemplateRetriever(TemplateRetriever):
             str: The wanted template
         """
         try:
-            return Template(self.table.loc[key])
+            result = self.table.loc[key]
+            if isinstance(result, (pd.Series, pd.DataFrame)):
+                raise KeyError(key)
+            return Template(result)
         except KeyError:
             return Template(default)
         except Exception:
@@ -403,8 +405,8 @@ class CsvTemplateRetriever(TemplateRetriever):
 class CentroidTemplateRetriever(CsvTemplateRetriever):
     """CentroidTemplateRetriever : TemplateRetriever for csv-centroids files.
 
-        Currently used by:
-            - mfire.utils.date : for describinng short periods
+    Currently used by:
+        - mfire.utils.date : for describinng short periods
     """
 
     def __init__(
@@ -428,12 +430,12 @@ class CentroidTemplateRetriever(CsvTemplateRetriever):
             self.weights = np.ones(len(self.table.index.names))
 
     @classmethod
-    def read_file(cls, filename: str, **kwargs) -> CentroidTemplateRetriever:
+    def read_file(cls, filename: Path, **kwargs) -> CentroidTemplateRetriever:
         """read_file : class method for instantiating an CentroidTemplateRetriever
         out of a file.
 
         Args:
-            filename (str): File's name
+            filename (Path): File's name
 
         Returns:
             CentroidTemplateRetriever: New cls object
@@ -455,7 +457,7 @@ class CentroidTemplateRetriever(CsvTemplateRetriever):
                 filename=filename,
                 exc_info=True,
             )
-            return cls({"filename": filename})
+            return cls({"filename": Path(filename)})
 
     def get(self, key: np.ndarray) -> str:
         """get: method used to retrieve a specific template identified by a key.
@@ -472,6 +474,66 @@ class CentroidTemplateRetriever(CsvTemplateRetriever):
         dist = np.sum(np.square(self.weights * (centroids - np.array(key))), axis=1)
         return Template(templates[dist.argmin()])
 
+    def get_by_dtw(
+        self,
+        key: np.ndarray,
+        pop_method: Optional[Union[str, Callable[Sequence, str]]] = "first",
+    ) -> str:
+        """get: method used to retrieve a specific template identified by a key.
+        It returns the template associated to the nearest centroid
+        of the given key with dtw method.
+
+        Args:
+            key (np.ndarray): Key for retrieving our template.
+            pop_method (Union[str, Callable[Sequence, str]], optional): Method to use
+                if the element retrieved with the key is not a string. It can be a
+                string among {"first", "last", "random"}, or it can be a function
+                which take the element as input and outputs a str.
+                Defaults to "first".
+
+        Returns:
+            dict: dict containing the min distance, the path, the template, the centroid
+        """
+        # il faut gérer le cas ou on a plusieurs distances mini egales
+
+        pop_methods = {
+            "first": lambda x: x[0],
+            "last": lambda x: x[-1],
+            "random": np.random.choice,
+        }
+        centroids = self.table.index.to_frame().values
+        datas = list()
+        for centroid in centroids:
+            data = [x for x in centroid if not pd.isnull(x)]
+            datas.append(data)
+        templates = self.table.values
+
+        distance_dict = dict()
+        for index, centroid in enumerate(datas):
+            distance_dict[tuple(centroid)] = {
+                "distance": dtw.distance_fast(
+                    np.array(key, dtype=np.double), np.array(centroid, dtype=np.double)
+                ),
+                "path": dtw.warping_path_fast(
+                    np.array(key, dtype=np.double), np.array(centroid, dtype=np.double)
+                ),
+                "template": Template(
+                    pop_methods.get(pop_method, pop_method)(templates[index].split("|"))
+                ),
+            }
+
+        d_min = {"distance": 9999, "path": None, "template": ""}
+
+        for centroid, result in distance_dict.items():
+            if result["distance"] > d_min["distance"]:
+                pass
+            else:
+                d_min = result
+                d_min["centroid"] = centroid
+                d_min["type"] = "general"
+
+        return d_min
+
     def __eq__(self, other: Any) -> bool:
         return super().__eq__(other) and np.all(self.weights == other.weights)
 
